# -*- coding: utf-8 -*-
"""Jaccard similarities and their p-values.

The code here represents a python implementation of the Jaccard package hosted
`here <https://github.com/ncchung/jaccard>`_ by N. Chung. Its citation follows.

Citation
--------
Chung, N., Miasojedow, B., Startek, M., and Gambin, A. "Jaccard/Tanimoto similarity test and estimation methods for biological presence-absence data" BMC Bioinformatics (2019) 20(Suppl 15): 644. https://doi.org/10.1186/s12859-019-3118-5
"""
import logging
from typing import Optional

import numpy as np
import pandas as pd

logger = logging.getLogger(__name__)


def similarity(
    x: np.ndarray,
    y: np.ndarray,
    center: bool = False,
    px: Optional[float] = None,
    py: Optional[float] = None,
) -> float:
    """Calculate Jaccard similarity.

    The vectors must be of the same length, and must be boolean.

    Parameters
    ----------
    x : np.ndarray
        A boolean array.
    y : np.ndarray
        A boolean array.
    center : bool, default=False  # noqa: DAR103
        Whether to center the score.
    px : Optional[float]
        The probability of success in x.
        If None, then px = x.mean()
    py : Optional[float]
        The probability of success in x.
        If None, then px = x.mean()

    Returns
    -------
    float
        The Jaccard similarity between the 2 vectors

    Raises
    ------
    IndexError
        If the vectors are not 1-d, or if the vectors are not the same length.
    TypeError
        If the vectors are not boolean.

    """
    if x.ndim != 1 or y.ndim != 1:
        logging.error(
            f"All vectors must be 1-d. ndims: {[x.ndim, y.ndim]}.", stack_info=True
        )
        raise IndexError
    if x.shape != y.shape:
        logging.error(
            f"All vectors must have the same length. shape: {[x.shape, y.shape]}",
            stack_info=True,
        )
        raise IndexError
    if x.dtype != bool or y.dtype != bool:
        logging.error(
            f"All vectors must be boolean. dtypes: {[x.dtype, y.dtype]}",
            stack_info=True,
        )
        raise TypeError

    if not px:
        logging.debug("px not provided. Calculating as mean of x.")
        px = x.mean()
        assert isinstance(px, float)
    if not py:
        logging.debug("py not provided. Calculating as mean of y.")
        py = y.mean()
        assert isinstance(py, float)

    intersect = (x & y).sum()
    union = x.sum() + y.sum() - intersect

    denominator = px + py - (px * py)
    if denominator == 0:
        logging.debug("Denominator is 0. Returning np.nan")
        return np.nan

    if union == 0:
        logging.debug("Union is 0. Calculating using px and py.")
        j = (px * py) / denominator
    else:
        j = intersect / union

    if center:
        logging.debug("Centering J.")
        return j - ((px * py) / denominator)
    else:
        return j


def distance(
    x: np.ndarray, y: np.ndarray, px: Optional[float] = None, py: Optional[float] = None
) -> float:
    """Calculate Jaccard distance.

    Classically defined as:

    Jdist = 1 - Jsimiliarity

    The vectors must be of the same length, and must be boolean.

    Note
    ----
    The centering method applied in Jaccard similarity is not
    applicable here, so it is not passed as an option.

    Parameters
    ----------
    x : np.ndarray
        A boolean array.
    y : np.ndarray
        A boolean array.
    px : Optional[float]
        The probability of success in x.
        If None, then px = x.mean()
    py : Optional[float]
        The probability of success in x.
        If None, then px = x.mean()

    Returns
    -------
    float
        The Jaccard distance between the 2 vectors
    """
    if not px:
        logging.debug("px not provided. Calculating as mean of x.")
        px = x.mean()
    if not py:
        logging.debug("py not provided. Calculating as mean of y.")
        py = y.mean()
    return 1 - similarity(x, y, center=False, px=px, py=py)


def bootstrap(
    x: np.ndarray,
    y: np.ndarray,
    px: Optional[float] = None,
    py: Optional[float] = None,
    n: int = 1000,
    seed: int = 42,
) -> pd.Series:
    """Use the bootstrap test to return a p-value.

    The p-value is defined as the fraction of values in the null statistic
    whose absolute value is greater than the absolute value of the observed
    statistic.

    Note
    ----
    Returning a series facilitates applications with pandas and groupby.

    Parameters
    ----------
    x : np.ndarray
        a boolean array.
    y : np.ndarray
        a boolean array.
    n : int
        the number of bootstrap repetitions to perform.
    px : Optional[float]
        The probability of success in x.
        If None, then px = x.mean()
    py : Optional[float]
        The probability of success in x.
        If None, then px = x.mean()
    seed : int
        the random seed to use for resampling.

    Returns
    -------
    Tuple[float, float]
        The Jaccard similarity and p_val.
    """
    if not px:
        logging.debug("px not provided. Calculating as mean of x.")
        px = x.mean()
    if not py:
        logging.debug("py not provided. Calculating as mean of y.")
        py = y.mean()

    j = similarity(x, y, center=False, px=px, py=py)
    if px == 1 or py == 1 or len(x) == x.sum() or len(y) == y.sum():
        logger.warning("Bootstrap is degenerate as at least one vector is all 1.")
        return pd.Series([j, 1], index=["J-sim", "p-val"])
    if px == 0 or py == 0 or x.sum() == 0 or y.sum() == 0:
        logger.warning("Bootstrap is degenerate as at least one vector is all 0.")
        return pd.Series([j, 1], index=["J-sim", "p-val"])

    j_obs = similarity(x, y, center=True, px=px, py=py)

    rng = np.random.default_rng(seed)
    vals = (  # pragma: no branch
        similarity(
            rng.choice(x, size=len(x), replace=True, shuffle=False),
            rng.choice(y, size=len(y), replace=True, shuffle=False),
            center=True,
        )
        for _ in range(n)
    )
    j_null = np.fromiter(vals, dtype=np.float32, count=n)
    p_val = (np.abs(j_null, dtype=np.float32, out=j_null) >= np.abs(j_obs)).sum() / n
    return pd.Series([j, p_val], index=["J-sim", "p-val"])



# -*- coding: utf-8 -*-
"""The root `lta` parser.

For a detailed description of the tool and how to use it,
please see the README.
A short overview is given here.

The location of the input data file is the first positional argument,
while the location to which the output files should be saved is the second positional.
The "0" threshold is specified with option ``-t/--threshold``.
This should be passed as a floating point number between 0 and 1, inclusive.
If it is not provided,
then a default of 0.3 is used.
The number of bootstrap replicates to use for estimating the p-values
can be specified with ``-b/--boot-reps``.
Generally, higher repetitions increases accuracy,
A deault of 1000 is used to provide a balance between speed and accuracy.

Many calculations are dependent on knowing where certain metadata is stored.
Namely, the experimental conditions (specified with ``--group``),
the compartment of origin (specified with ``--compartment``),
the sample ID (specified with ``--sample-id``),
and the lipidomics mode (specifed with ``--mode``).
If these are not passed,
then they default to "Group", "Compartment", "SampleID", and "Mode" respectively.

We also have to know how many rows of metadata there are.
This is specified with ``--n-rows-metadata``,
and defaults to 11 if not given.

The error-normalised fold change (ENFC) calculation must know the labels for
experimental and control group.
Without this knowledge,
the concept of fold change is meaningless.
To specify, pass ``--control control``.
Every condition specified in ``Group`` will then be divided by ``control``
to calculate the ENFC for all conditions.

Finally,
all options *can be* passed in an config file.
This file expects values of the format ``option=value``.
By default,
the CLI looks for ``lta_conf.txt``,
though any file can be specified with the ``-c/--config`` flag.
Values in the config file override the defaults,
and values passed to the CLI override the config file.

Like all good CLIs,
``-V`` returns the version while ``-h/--help`` returns help.

Attributes
----------
lta_parser : argparse.ArgumentParser
    The argument parser for the root command.
"""
from pathlib import Path

import configargparse

from lta import __version__
from lta.commands.run import run
from lta.helpers.custom_types import FloatRange

lta_parser = configargparse.ArgumentParser(
    prog="lta",
    description="Lipid Traffic Analysis",
    allow_abbrev=False,
    add_config_file_help=True,
    default_config_files=["lta_conf.txt"],
    config_arg_is_required=False,
    ignore_unknown_config_file_keys=True,
    formatter_class=configargparse.ArgumentDefaultsRawHelpFormatter,
)

lta_parser.add_argument(
    "-c",
    "--config",
    required=False,
    is_config_file=True,
    help="Config file location.",
)

lta_parser.add_argument(
    "file",
    type=Path,
    help="Location of the input data csv file.",
)

lta_parser.add_argument(
    "output",
    type=Path,
    help="Where to write output file.",
)

# The type ignore silence mypy.
# Since we are checking, not iterating, a proper __iter__ is unnecessary
lta_parser.add_argument(
    "-t",
    "--threshold",
    type=float,
    choices=FloatRange(0, 1),  # type: ignore
    default=0.3,
    help="The '0' threshold",
)

lta_parser.add_argument(
    "-b",
    "--boot-reps",
    type=int,
    default=1000,
    help="Number of bootstrap repetitions",
)

lta_parser.add_argument(
    "-n",
    "--n-rows-metadata",
    type=int,
    default=11,
    help="Number of rows in column metadata",
)

lta_parser.add_argument(
    "--group",
    type=str,
    default="Group",
    help="Metadata label for experimental conditions",
)

lta_parser.add_argument(
    "--control",
    type=str,
    default="control",
    help="Control group for fold-change",
)

lta_parser.add_argument(
    "--compartment",
    type=str,
    default="Compartment",
    help="Metadata label for sample compartment",
)

lta_parser.add_argument(
    "--mode",
    type=str,
    default="Mode",
    help="Metadata label for lipidomics mode",
)

lta_parser.add_argument(
    "--sample-id",
    type=str,
    default="SampleID",
    help="Metadata label for Sample IDs",
)

lta_parser.add_argument(
    "-V",
    "--version",
    action="version",
    version=f"LTA v{__version__}",
    help="Display version information and exit.",
)

lta_parser.add_argument(
    "-v",
    "--verbose",
    action="count",
    default=0,
    help="Increase verbosity.",
)

lta_parser.add_argument(
    "-l",
    "--logfile",
    action="append",
    help="Location of logfile. May also be 'term' for Std.Out.",
)

lta_parser.set_defaults(func=run)



# -*- coding: utf-8 -*-
"""A dataclass that allows for an object oriented pipeline."""
import itertools
import logging
from dataclasses import dataclass
from pathlib import Path
from typing import Dict

import pandas as pd

import lta.helpers.data_handling as dh
import lta.helpers.jaccard as jac

logger = logging.getLogger(__name__)


@dataclass
class Pipeline:
    """The Lipid Traffic Analysis pipeline.

    Attributes
    ----------
    file : Path
        The path to the combined data input file.
    output : Path
        Where to save the results.
    n_rows_metadata : int
        The number of rows to treat as column metadata.
        As Python is 0-indexed, passing ``11`` will read in rows ``0-10``.
    level : str
        Metadata location of experimental conditions.
    control : str
        Value within self.level that represents the control condition.
    compartment : str
        Metadata location of sample tissue compartment.
    mode : str
        Metadata location of lipidomics mode.
    sample_id : str
        Metadata location of sample IDs.
    thresh : float
        The fraction of samples that are 0 above which a lipid will be called 0 for a compartment.
    n : int
        Number of bootstrap replicates.
    """

    file: Path
    output: Path
    n_rows_metadata: int
    level: str
    control: str
    compartment: str
    mode: str
    sample_id: str
    thresh: float
    n: int

    def __post_init__(self) -> None:
        """Post-process parameters.

        The magic of DataClasses!
        The post-init method allows for much of the processing normally required.
        Several things happen here.
        The file is read into a dataframe,
        which is split by ``Mode``,
        and stored twice as binary data (for Switch Analysis)
        and filtered counts (for ENFC Analysis).
        Any sample which contains all 0-values is dropped.
        Should the file not exist,
        not contain any data,
        or be a directory,
        then the appropriate errors are logged.
        Also, output folder is created if it does not exist already.

        Raises
        ------
        FileNotFoundError
            If self.file does not exist.
        IsADirectoryError
            If self.file is a directory.
        pd.errors.EmptyDataError
            If there is no data in self.file.
        """
        try:
            logger.debug(f"Reading data from {self.file}...")
            data = dh.construct_df(
                self.file,
                self.n_rows_metadata,
                [self.mode, self.level, self.compartment, self.sample_id],
                index_names=["Lipid", "Category", "m/z"],
                index_col=[0, 1, 2],
                header=None,
            ).pipe(
                lambda x: x.loc[:, x.any()]
            )  # Drop all-0 samples
        except FileNotFoundError:
            logger.exception(f"{self.file} does not exist. A full traceback follows...")
            raise
        except IsADirectoryError:
            logger.exception(f"{self.file} is a directory. A full traceback follows...")
            raise
        except pd.errors.EmptyDataError:
            logger.exception(
                f"{self.file} contains no data. A full traceback follows..."
            )
            raise
        logger.debug("Binarizing data...")
        self.binary = {
            group: dh.not_zero(
                df,
                axis="columns",
                level=self.level,
                compartment=self.compartment,
                thresh=self.thresh,
            )
            for group, df in data.groupby(axis="columns", level=self.mode)
        }
        logger.debug("Filtering data...")
        self.filtered = {
            group: df.loc[self.binary[group].index, :]
            for group, df in data.groupby(axis="columns", level=self.mode)
        }
        Path(self.output, "enfc").mkdir(exist_ok=True, parents=True)
        Path(self.output, "jaccard").mkdir(exist_ok=True, parents=True)

        conditions = [
            df.columns.get_level_values(self.level).unique()
            for df in self.filtered.values()
        ]
        self.conditions = [
            val for mode in conditions for val in mode if val != self.control
        ]

    def _calculate_enfc(self) -> Dict[str, Dict[str, pd.DataFrame]]:
        """Calculate error-normalised fold change.

        Calculates the ENFC for each compartment across modes.
        For fold change to be meaningful,
        order must be specified.
        This will report fold-change as
        ``condition / self.control`` for all conditions
        except control within self.value.

        Returns
        -------
        Dict[str, Dict[str, pd.DataFrame]]
            Top level key is the experimental condition,
            mapped to a dictionary of modes and ENFC results
        """
        logger.info("Calculating ENFC...")
        enfc = {
            group: {
                mode: df.groupby(axis="columns", level=self.compartment).agg(
                    dh.enfc,
                    axis="columns",
                    level=self.level,
                    order=(group, self.control),
                )
                for mode, df in self.filtered.items()
            }
            for group in self.conditions
        }
        return enfc

    def _get_a_lipids(self) -> Dict[str, pd.DataFrame]:
        """Extract A-lipids from the dataset.

        Any compartment where more than self.thresh of the samples are 0
        is considered a total 0 for that lipid.
        Lipids that are non-0 for all compartments in any Phenotype
        are considered A-lipids.

        Returns
        -------
        Dict[str, pd.DataFrame]
            Key is mode, value is table of A-lipids
        """
        logger.info("Calculating A-lipids...")
        results = {
            f"a_{mode}": (
                df.groupby(axis="columns", level=self.level)
                .all()
                .pipe(lambda x: x.loc[x.any(axis="columns"), :])
            )
            for mode, df in self.binary.items()
        }
        return results

    def _get_b_lipids(self, picky: bool = True) -> Dict[str, pd.DataFrame]:
        """Extract B-lipids from the dataset.

        Any compartment where more than self.thresh of the samples are 0
        is considered a total 0 for that lipid.
        Lipids that are non-0 for any pair of compartments within any Phenotype
        are considered B-lipids.

        Notes
        -----
        By definition,
        all A-lipids will also be B-lipids for every pair of compartments.
        These are referred to as B-consistent,
        or Bc for short.
        Those B-lipids that are not A-lipids are also known as B-picky,
        or Bp for short.
        Which set is calculated is controlled by the boolean flag ``picky``.

        Parameters
        ----------
        picky : bool
            default=True
            If true, do **not** consider lipids that are also A-lipids

        Returns
        -------
        Dict[str, pd.DataFrame]
            Keys are the compartment pair and mode.
            Values are the table of B-lipids for that grouping.

        Raises
        ------
        AttributeError
            If A-lipids have not been previously calculated
        """
        # drop or keep a-lipids, depending on picky
        try:
            a_lip = {mode: df.index for mode, df in self.a_lipids.items()}
        except AttributeError:
            logger.exception("You must find A-lipids before B-lipids.")
            raise
        else:
            # This assumes that self.binary and a_lip will have the same keys
            # Which is definitely True
            if picky:
                data = {
                    mode: df.drop(index=a_lip[f"a_{mode}"])
                    for mode, df in self.binary.items()
                }
                subtype = "p"
            else:
                data = {
                    mode: df.loc[a_lip[f"a_{mode}"], :]
                    for mode, df in self.binary.items()
                }
                subtype = "c"

        logger.info(f"Calculating B{subtype}-lipids...")
        results = {}
        for mode, df in data.items():
            logger.debug(f"Calculating B{subtype}-lipids for {mode}...")
            compartments = df.columns.get_level_values(self.compartment)
            pairs = itertools.combinations(compartments.unique(), 2)
            for group in pairs:
                logger.debug(f"Calculating B{subtype}-lipids for {group}...")
                unified = (
                    df.loc[:, compartments.isin(group)]
                    .groupby(axis="columns", level=self.level)
                    .all()
                    .pipe(lambda x: x.loc[x.any(axis="columns"), :])
                )
                pairing = "_".join([x.upper() for x in group])
                results[f"b{subtype}_{pairing}_{mode}"] = unified
        return results

    def _get_n_lipids(self, n: int) -> Dict[str, pd.DataFrame]:
        r"""Extract N-lipids from the dataset.

        Any compartment where more than self.thresh of the samples are 0
        is considered a total 0 for that lipid.
        Lipids that are non-0 for ``n`` compartments in any Phenotype
        are considered N-lipids.

        Notes
        -----
        For historical consistency,
        N1-lipids (*ie* those found in only 1 compartment) are called U-lipids
        as they are **U**\nique.
        Also N2-lipids are not the same as B-lipids.
        A B-lipid could occur in multiple pairs of compartment,
        while N2-lipids must only occur in 1.

        Parameters
        ----------
        n : int
            The number of compartments to limit the search to

        Returns
        -------
        Dict[str, pd.DataFrame]
            Keys are the compartment group and mode.
            Values are the table of N-lipids for that grouping.
        """
        logger.info(f"Calculating N{n}-lipids...")
        results = {}
        for mode, df in self.binary.items():
            logger.debug(f"Calculating N{n}-lipids for {mode}...")
            compartments = df.columns.get_level_values(self.compartment)
            # Mask required to prevent dropping levels
            # The initial check must be done with all compartments (only n)...
            mask = (
                df.groupby(axis="columns", level=self.compartment)
                .any()
                .sum(axis="columns")
                == n
            )

            data = [
                (group, df.loc[mask, compartments.isin(group)])
                for group in itertools.combinations(compartments.unique(), n)
            ]
            logger.debug(
                f"N{n} compartment groups before filtering: {[group for group, _ in data]}"
            )
            # ...which necessitates a second check to drop those that are not
            # Again, mask necessary for keeping info
            # Also, we only care for groups with lipids
            masks = [
                df.groupby(axis="columns", level=self.compartment)
                .any()
                .sum(axis="columns")
                == n
                for _, df in data
            ]
            data = [
                (group, df.loc[mask, :].groupby(axis="columns", level=self.level).all())
                for (group, df), mask in zip(data, masks)
                if mask.sum() != 0
            ]
            logger.debug(
                f"N{n} compartment groups after filtering: {[group for group, _ in data]}"
            )
            for (compartments, df) in data:
                n_type = "u" if n == 1 else f"n{n}"
                group = "_".join([x.upper() for x in compartments])
                results[f"{n_type}_{group}_{mode}"] = df
        return results

    def _jaccard(
        self, data: Dict[str, pd.DataFrame], group: str
    ) -> Dict[str, Dict[str, pd.DataFrame]]:
        """Calculate jaccard similarity and p-values.

        This takes a dictionary of data.
        As the output of each group of lipids will be structured as such,
        it should be called per lipid group.

        Notes
        -----
        The P-values are calculated using a bootstrap approach on a centered Jaccard similarity.

        Parameters
        ----------
        data : Dict[str, pd.DataFrame]
            A dictionary of modes and lipid data
        group : str
            Which lipids are being checke.
            This is used by logging only.

        Returns
        -------
        Dict[str, Dict[str, pd.DataFrame]]
            Keys are the compartment group and mode.
            Values are the table of Jaccard similarity and p-values.
        """
        logger.info(f"Calculating Jaccard similarity for {group}...")
        jaccard = {
            group: {
                mode: lipids.loc[:, [group, self.control]]
                .groupby(axis="index", level="Category")
                .filter(lambda df: df.sum().any())
                .groupby(axis="index", level="Category")
                .apply(
                    lambda x: jac.bootstrap(
                        x.loc[:, group], x.loc[:, self.control], self.n
                    )
                )
                for mode, lipids in data.items()
            }
            for group in self.conditions
        }
        return jaccard

    def run(self) -> None:
        """Run the full LTA pipeline.

        This:

        #. Calculates error-normalised fold change for all conditions relatve to control
        #. Finds A-lipids and Jaccard distances.
        #. Finds U-lipids and Jaccard distances.
        #. Finds B-lipids (both picky and consistent) and Jaccard distances.
        #. Finds N2-lipids and Jaccard distances.
        #. Writes combined results.
        """
        logger.debug("Generating ENFC summary files...")
        self.enfcs = self._calculate_enfc()
        for group, data in self.enfcs.items():
            pd.concat(data, axis="columns").to_csv(
                self.output
                / "enfc"
                / f"{group}_by_{self.control}_individual_lipids.csv"
            )

        logger.debug("Generating class ENFC summary files...")
        self.filtered = {
            mode: df.groupby(axis="index", level="Category").sum()
            for mode, df in self.filtered.items()
        }
        self.enfcs = self._calculate_enfc()
        for group, data in self.enfcs.items():
            pd.concat(data, axis="columns").to_csv(
                self.output / "enfc" / f"{group}_by_{self.control}_lipid_classes.csv"
            )

        self.a_lipids = self._get_a_lipids()
        self.a_jaccard = self._jaccard(self.a_lipids, "A-lipids")

        self.u_lipids = self._get_n_lipids(1)
        self.u_jaccard = self._jaccard(self.u_lipids, "U-lipids")

        self.bc_lipids = self._get_b_lipids(picky=False)
        self.bc_jaccard = self._jaccard(self.bc_lipids, "Bc-lipids")

        self.bp_lipids = self._get_b_lipids(picky=True)
        self.bp_jaccard = self._jaccard(self.bp_lipids, "Bp-lipids")

        self.n2_lipids = self._get_n_lipids(2)
        self.n2_jaccard = self._jaccard(self.n2_lipids, "N2-lipids")

        logger.debug("Generating Switch Analysis (lipid-type) summary files...")
        summary = pd.concat(
            {
                **self.a_lipids,
                **self.bc_lipids,
                **self.bp_lipids,
                **self.n2_lipids,
                **self.u_lipids,
            },
            axis="columns",
        ).fillna(False)
        summary.columns.names = ["type_compartment_mode", "Phenotype"]
        summary.to_csv(self.output / "switch_individual_lipids.csv")
        summary.groupby(axis="index", level="Category").sum().to_csv(
            self.output / "switch_lipid_classes.csv"
        )

        logger.debug("Generating Jaccard distance summary files...")
        for group in self.conditions:
            jaccard = pd.concat(
                {
                    **self.a_jaccard[group],
                    **self.bc_jaccard[group],
                    **self.bp_jaccard[group],
                    **self.n2_jaccard[group],
                    **self.u_jaccard[group],
                },
                axis="columns",
            )
            jaccard.columns.names = ["type_compartment_mode", "Metrics"]
            jaccard.to_csv(
                self.output
                / "jaccard"
                / f"{group}_to_{self.control}_jaccard_similarity.csv"
            )



# -*- coding: utf-8 -*-
"""LTA: A Python CLI for Lipid Traffic Analysis.

Attributes
----------
__version__ : str
    The version number,
    specified in the form 'major.minor.patch'
"""
__version__ = "3.0.0"



# -*- coding: utf-8 -*-
"""A simple sample function for the CLI."""
import logging

import configargparse

from lta.helpers.pipeline import Pipeline

logger = logging.getLogger(__name__)


def run(args: configargparse.Namespace) -> None:
    """Initialise and run the LTA pipeline.

    This function constructs a ``lta.helpers.pipeline.Pipeline`` instance from
    the ``configargparser.Namespace``,
    and then calls the ``run`` method.
    Nothing more, nothing less.
    The rest of the magic is handled by the pipeline.

    Parameters
    ----------
    args: configargparse.Namespace
        The passed args.
    """
    logger.debug("Instantiating pipeline instance.")
    pl = Pipeline(
        args.file,
        args.output,
        args.n_rows_metadata,
        args.group,
        args.control,
        args.compartment,
        args.mode,
        args.sample_id,
        args.threshold,
        args.boot_reps,
    )
    logger.debug("Running pipeline.")
    pl.run()



# -*- coding: utf-8 -*-
"""Provide entry point for CLI."""
import logging
from datetime import datetime
from pathlib import Path
from typing import Optional, cast

import configargparse

from lta.parser import lta_parser


def main(args: Optional[configargparse.Namespace] = None) -> None:
    """Provide entry point for CLI.

    This also configures the logger to use for the run.
    The ``-v`` flag can be passed multiple times,
    increasing verbosity from errors to debug.
    Additionally,
    file handlers are set up for ``--logfile``,
    unless the passed file is ``term``,
    in which case a stream handler is used.

    Parameters
    ----------
    args : Optional[configargparse.Namespace]
        If not specified, then the command line inputs are read.
        This is the *intended* behaviour,
        as this option only exists to allow for testing.
    """
    if args is None:
        args = lta_parser.parse_args()

    # Get verbosity level
    verbosity = {
        0: logging.ERROR,
        1: logging.WARNING,
        2: logging.INFO,
        3: logging.DEBUG,
    }
    level = verbosity.get(args.verbose, logging.DEBUG)

    # Configure formatter
    formatter = logging.Formatter(
        "{asctime} :: {levelname} :: {name} :: {message}", style="{"
    )

    # Configure handlers
    # Append does not overwrite defaults...
    # Dashes are used instead of my usual colons for time,
    # As they are Windows incompatible.
    if not args.logfile:
        logs = Path("logs")
        logs.mkdir(parents=False, exist_ok=True)
        now = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
        args.logfile = [logs / f"{now}.log"]

    handlers = [
        cast(logging.Handler, logging.StreamHandler())
        if log == "term"
        else cast(logging.Handler, logging.FileHandler(log))
        for log in args.logfile
    ]
    for h in handlers:
        h.setLevel(level)
        h.setFormatter(formatter)

    # Configure root logger
    logging.basicConfig(level=level, handlers=handlers)
    logger = logging.getLogger(__name__)

    logger.log(
        45, f"Running LTA with the following parameters:\n{lta_parser.format_values()}"
    )
    args.func(args)



# -*- coding: utf-8 -*-
"""Helper functions for handling data.

In order to make the pipeline from ``lta.helpers.pipeline`` more testable,
most of its constructor and processing functions have been moved here.
This makes the code more atomic -
and, thus, testable -
by removing it from the harder to test context of the object.
"""
import logging
from pathlib import Path
from typing import Any, List, Literal, Optional, Tuple

import numpy as np
import pandas as pd

logger = logging.getLogger(__name__)


def construct_df(
    file: Path,
    n_rows: int,
    metadata: List[str],
    index_names: Optional[List[str]] = None,
    **kwargs: Any,
) -> pd.DataFrame:
    """Construct a dataframe from the given path.

    A light wrapper to handing the reading of complex metadata.
    It reads the whole dataframe,
    then sices of ``n_rows`` to treat as metadata.
    Any undesired rows are dropped,
    then metadata is added as a multiindex onto the raw data,
    less those rows that were metadata.

    Any row metadata can be retained by specifying the
    ``index_col`` kwarg to ``pd.read_csv``.

    Parameters
    ----------
    file : Path
        The path to the data file
    n_rows : int
        The number of rows in the column metadata
    metadata : List[str]
        The metadata rows to include
    index_names : Optional[List[str]]
        Names for the data frame multi-index
    **kwargs : Any
        Further argument passed to ``pd.read_csv``

    Returns
    -------
    pd.DataFrame
        The created dataframe
    """
    counts: pd.DataFrame = pd.read_csv(file, **kwargs)

    # Retrieve column metadata
    cols = counts.iloc[:n_rows, :].copy()
    cols.index = cols.index.droplevel([0, 1])
    cols = cols.loc[metadata, :]

    # Create final dataframe
    data = counts.iloc[n_rows:, :].copy()
    data.columns = pd.MultiIndex.from_frame(cols.transpose())
    if index_names:
        data.index.names = index_names

    # Force numeric types
    data = data.apply(pd.to_numeric)

    return data


def not_zero(
    df: pd.DataFrame,
    axis: Literal["index", "columns"],
    level: str,
    compartment: str,
    thresh: float,
) -> pd.DataFrame:
    """Mark any group that has more than ``thresh`` fraction 0 as 0.

    Given a counts matrix, mark all counts that are zero.
    Then, groupby the level of the specified multiindex,
    and mark a whole group as 0 if there are more than
    ``thresh * len(group)`` 0s.
    Finally,
    any lipid that is all 0 is dropped on the specified ``axis``

    Grouping occurs across compartment and level in that order.

    Parameters
    ----------
    df : pd.DataFrame
        The lipid data to convert to boolean
    axis : Literal['index', 'columns']
        Which multiindex to consider
    level : str
        The level of the multiindex to groupby
    compartment : str
        The level of the multiindex containing compartment sample
    thresh : float
        The fraction of samples above which a group is said to be 0

    Returns
    -------
    pd.DataFrame
        The processed data.
    """
    df = (
        (df == 0)
        .groupby(axis=axis, level=[compartment, level])
        .agg(lambda x: x.sum() <= (thresh * len(x)))
    )
    if axis == "index":
        df = df.loc[:, df.any(axis=axis)]
    if axis == "columns":
        df = df.loc[df.any(axis=axis), :]
    return df


def enfc(
    df: pd.DataFrame,
    axis: Literal["index", "columns"],
    level: str,
    order: Optional[Tuple[str, str]] = None,
) -> pd.DataFrame:
    """Calculate the error normalised fold change for a dataframe.

    ENFC is defined as the log foldchange of lipid levels divide by the propagated error.
    The mean and standard deviation are calculated on the groups in ``level``,
    before applying the above calculation.

    Notes
    -----
    By definition,
    fold change requires an experimental and control group,
    othewise the notion of up- or down-regulated makes no sense.
    This function assumes these groups are named "experimental" and "control",
    respectively,
    though alternatives can be passed to the ``order`` argument.

    Additionally,
    the notion of fold change requires both samples to be non-0.
    When either x/0, 0/x, or 0/0 occurs during the FC calculation,
    these values are returned as NaN.

    Parameters
    ----------
    df : pd.DataFrame
        The dataframe containing lipid expression values
    axis : Literal['index', 'columns']
        Which multiindex to consider
    level : str
        The level of the multiindex containing experimental conditions
    order : Optional[Tuple[str, str]]
        default=('experimental', 'control')
        The names of the conditions to compare.
        Fold change will be ``order[0] / order[1]``.

    Returns
    -------
    pd.DataFrame
        The processed data.
    """
    if not order:
        logging.debug("Order not passed. Defaulting to ('experimental', 'control')")
        order = ("experimental", "control")
    mean = df.groupby(axis=axis, level=level).mean()
    logging.debug(f"Grouping/filtering on {axis}.")
    # Replace inf (x/0) with NaN
    # Replace 0 (0/x) with NaN
    # 0/0 is already NaN
    if axis == "index":
        logfc = np.log10(
            mean.loc[order[0], :]
            .div(mean.loc[order[1], :])
            .replace([np.inf, -np.inf, 0], np.NAN)
        )
    else:
        logfc = np.log10(
            mean.loc[:, order[0]]
            .div(mean.loc[:, order[1]])
            .replace([np.inf, -np.inf, 0], np.NAN)
        )
    error = (
        df.groupby(axis=axis, level=level).std().pow(2).sum(axis=axis).div(2).pow(0.5)
    )
    return logfc.div(error)



# -*- coding: utf-8 -*-
"""Custom types for controlling argparse selections."""


class FloatRange:
    """A custom range that supports floating point numbers.

    This does **not** implement all the features of the default `range`.
    Namely, it doesn't support an interval as this does not makes sense
    in the context of floats, as they are an infinitely sub-dividable set.

    Its principle use is to provide a custom container to argparse to restrict
    the values a float might be.
    """

    def __init__(self, start: float, end: float) -> None:
        """Initialise the object.

        Parameters
        ----------
        start : float
            The start of the range, inclusive.
        end : float
            The end of the range, inclusive.
        """
        self.start = start
        self.end = end

    def __repr__(self) -> str:
        """Return the representation of the object.

        Returns
        -------
        str
            The FloatRange representation.
        """
        return f"[{self.start}, {self.end}]"

    def __contains__(self, x: float) -> bool:
        """Test if the FloatRange contains a value.

        Parameters
        ----------
        x : float
            The value to check.

        Returns
        -------
        bool
            True if x in FloatRange, False otherwise.
        """
        return self.start <= x <= self.end

    def __iter__(self) -> "FloatRange":  # type: ignore
        """Iterate over the object.

        This is a bit of cheeky magic to force argparse to recognise the container,
        which is why the typing is ignored.
        It is not designed to actually be iterated over,
        as it will perpetually yield self.

        Yields
        ------
        FloatRange
            Self, yielded indefinitely.
        """
        yield self



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['lta', 'lta.commands', 'lta.helpers']

package_data = \
{'': ['*']}

install_requires = \
['ConfigArgParse>=1.5,<2.0', 'numpy>=1.22.2,<2.0.0', 'pandas>=1.3,<2.0']

entry_points = \
{'console_scripts': ['lta = lta.cli:main']}

setup_kwargs = {
    'name': 'lipidta',
    'version': '3.0.0',
    'description': 'Lipid Traffic Analysis',
    'long_description': '# Lipid Traffic Analysis\n\n[![MIT License](https://img.shields.io/badge/License-MIT-blue.svg)](https://opensource.org/licenses/MIT)\n[![PyPI Version](https://img.shields.io/pypi/v/lipidta)](https://pypi.org/project/gtexquery/)\n[![Python Versions](https://shields.io/pypi/pyversions/lipidta)](https://shields.io/pypi/pyversions/lipidta)\n[![CI/CD](https://github.com/IMS-Bio2Core-Facility/lta/actions/workflows/cicd.yaml/badge.svg)](https://github.com/IMS-Bio2Core-Facility/lta/actions/workflows/cicd.yaml)\n[![codecov](https://codecov.io/gh/IMS-Bio2Core-Facility/lta/branch/main/graph/badge.svg?token=2TGYX69U3N)](https://codecov.io/gh/IMS-Bio2Core-Facility/lta)\n[![Documentation Status](https://readthedocs.org/projects/lta/badge/?version=latest)](https://lta.readthedocs.io/en/latest/?badge=latest)\n[![Project Status: Active](https://www.repostatus.org/badges/latest/active.svg)](https://www.repostatus.org/#active)\n[![Codestyle: Black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![Imports: isort](https://img.shields.io/badge/%20imports-isort-%231674b1?style=flat&labelColor=ef8336)](https://pycqa.github.io/isort/)\n\n**aka LTA, aka LipidTA**\n\nA python commandline interface for analysing lipidomics data.\n\nThe source code lives on [github][github].\n\nThe documentation lives at [ReadTheDocs][readthedocs].\n\nThe project can be installed from [PyPI][pypi].\n\n## Abstract\n\nLipid Traffic Analysis (LTA) is a tool for using lipidomics data to test hypotheses about how metabolism is controlled.\nLipidomics data from several, metabolically connected tissues from control and experimental groups\ncan be used to plot the spatial or temporal distribution of lipids.\nThese distributions identify where changes in lipid metabolism occur and in which lipid pathways,\nindicating the locus and biochemical alterations that occur in a given group.\nLTA was conceived in two parts.\nOne is an Abundance Analysis,\nin which the error-normalised fold change (ENFC) for the control and given group group is calculated.\nBecause the ratio of the control and experimental values is scaled by the error,\nthe ENFCs are easy to plot and compare between compartments.\nThe second part is a Switch Analysis.\nThis computes the presence of variables across the network.\nCurrent development is focused on developing the technique for complex networks and on the rate of lipid transport.\n\n### On Version Numbers\n\n```{admonition} TLDR\nPrior to version 3,\nthe Python implementation version **does not** correlate to the academic versions.\nFrom version 3,\nthe Python and academic versions are the same.\n```\n\nLTA was developed initially in R,\nproducing [LTA v1.0][lta_v1].\nThis was used for the LTA in the first study to use the analysis,\nbased on the hypothesis that differing carbohydrate intakes by fathers\ninfluenced their children\'s and grandchildren\'s lipid metabolism.\nSee [Furse et al 2021][paper_1].\nThe code was developed further in R to improve convenience for the user,\nand features such as the 0s threshold\n(the number of samples in which there is a non-zero value for abundance for it to count as present)\nwas added for LTA v2.3.\nThis version has been used in three subsequent studies.\nPlease see [here][paper_2],\n[here][paper_3],\nand [here][paper_4].\nA desire to add more features and move to a more stable platform led us to move to Python for LTA v3.0.\n\nThe Python implementation of LTA has strictly used automated semantic versioning over the course of its development.\nThus,\nprior to version 3,\n**the version numbers of the Python inplementation and past R implementations have no correlation.**\nBy convenient coincidence,\nthe semantic versioning release of LipidTA v3 brings\nthe python versioning and the academic versioning in-line.\n\n## Using LTA from the command line\n\n### Installation\n\nAny installation of this tool requires Python.\nHow to install Python correctly is beyond this scope,\nbut there are some excellent resources available\n[here][realpython]\nand [here][psf].\nIf you are planning to have multiple versions of python installed,\nyou might want to consider [pyenv][pyenv].\n\n```{warning}\nWe only support Python {math}`>=` 3.7.1\n```\n\n#### Installing from PyPI\n\nThis is the most straightforward way to set up the tool.\nWhen installing from PyPI,\nwe strongly reccomend using a virtual environment.\nThere are many ways to do this!\nIf you already have a preferred method -\nI use [pipx][pipx] for command line tools -\nfeel free to use that.\nOtherwise,\nuse the builtin Python module [venv][venv].\nThe exact instructions are OS-specific and detailed at the above link.\nInstructions for installing the most recent version of LTA on MacOS are given below:\n\n```shell\n# Make a directory for the project\nmkdir lta && cd lta\n# Create the virtual environment\npython3 -m venv .venv\n# Activate the environment\nsource .venv/bin/activate\n# Install lta\npip install -U LipidTA\n```\n\n```{note}\nOur pip package is `LipidTA`.\nUnfortunately,\n`lta` was "too similar to existing package names",\nso PyPi wouldn\'t let us use it.\n```\n\nIf you want to install a specific version,\nthen change the last line in the previous code block to:\n\n```shell\npip install LipidTA==1.0.0\n```\n\nreplacing the version number with the version number you want.\nA list of all released versions can be found at our [tags][tags].\n\n#### Installing from Source\n\n```{important}\nMost users **will not need** these instructions.\n```\n\nIf you need to customise the code in some manner,\nyou\'ll need to install from source.\nTo do that,\neither clone the repository from github,\nor download one of our releases.\nFor full instructions,\nplease see our guide on [contributing](./contributing.md).\n\n(data)=\n\n### The Data\n\nThe input should be a csv containing the lipidomics results.\nThough we strive to be as flexible as possible,\nwe must make some assumptions about the data to be able to use it.\nFirstly,\nthe first 3 columns must be the multiindex for the lipids,\nand include the lipid name,\ncategory,\nand m/z, respectively.\nSecondly,\nthe values must be numeric.\n\nThe analysis depends on a number of key metadata variables,\nnamely:\n\n- Mode: the lipidomics mode\n- Sample ID: a unique sample identifier\n- Group: the experimental condition\n- Compartment: the comartment of origin for the sample\n\nThese rows should be in the first {math}`n` rows of your data file,\nwhere n is specified with the option `--n-rows-metadata`.\nYou can name these metadata rows whatever you want in the data file,\nand tell `lta` where to find them with the appropriate flags.\nWe\'ve done our best to choose sane defaults\n(please see the section on [default values](defaults)),\nand you can see the section on [customising your run](customising)\nfor more details.\nHowever,\nif these data are not present,\nthe tool will not run,\nas the analysis only makes sense in the context of these variables.\n\n```{important}\nShould we make any changes to data format expectations,\nit will be well documented and will only occur in a major/breaking releas.\n```\n\n### Running the analysis\n\nOnce you\'ve installed the tool and activated your virtual environment,\nrunning the analysis **can** be as simple as:\n\n```shell\nlta data.csv results\n```\n\nThe first argument is path to the combined input file.\nIf the file doesn\'t exist,\nis a directory,\nor doesn\'t contain any data files,\nthe command will error with an apropriate message.\nThe secont argument identifies a folder in which the results will be saved.\nIt will be created if it doesn\'t exist.\n\n```{important}\nTo help get you up and running,\na minimum dataset and configuration file are provided [here][examples].\n```\n\n(defaults)=\n\n#### The Defaults\n\nTo keep it as simple as above,\nyou will need to use the defaults,\noutlined in the below table:\n\n| Parameter | Description | Default |\n| :--- | :---: | ---: |\n| threshold | The fraction of samples in which a lipid is 0 before it is dropped | 0.3 |\n| boot-reps | The number of bootstrap repetitions used to calculate probability | 1000 |\n| n-rows-metadata | The number of rows of metadata at the beginning of the data file| 11 |\n| group | The metadata row containing experimental conditions | Group |\n| control | The "control" condition, used as reference for fold change | control |\n| comartment | The metadata row containing the compartment of origin for each sample | Compartment |\n| mode | The metadata row containing the lipidomics mode | Mode |\n| sample-id | The metadata ro containing unique sample identifiers | SampleID |\n\nDon\'t worry if it looks intimidating!\nYou can check out the section on [customising your run](customising)\nfor further details,\nand help can always be found at our [documentation][readthedocs]\nor from the command line with:\n\n```shell\nlta -h\n```\n\n(customising)=\n\n#### Customising\n\nWhile it can be that simple,\nyou\'ll likely have to customise some options for your run.\nIn that case,\nit will likely look a bit more like:\n\n```shell\nlta --n-rows-metadata 11 \\\n--group Group \\\n--control lean \\\n--compartment Compartment \\\n--sample-id mouse\n```\n\nAlternatively,\nyou might prefer to use a configuration file to keep things simple.\nIn that case,\nsee the section on [configuration](configuration)\nfor more information.\n\nThere are a few options that can be customised for any given run.\nThe statistics are calculated using a bootstrapping approach,\nwhich (by definition) involves repeated replicates.\nTo control the number of replicates,\npass the ``-b/--boot-reps`` flag with a number.\nGenerally, more reps improves the accuracy of the estimates,\nthough I find little improvement beyond 20,000 reps.\n1000 (the default number) seems to provide a good balance between speed and accuracy.\n\nA critical step of the analysis is binarizing the lipid expression.\nA lipid is classed as 0 in a compartment/condition if\nthe lipid is **not** detected in more than a particular fraction of samples.\nThe default values is 0.2 (one-fifth of the samples).\nIf you want to change it,\npass the ``-t/--threshold`` flag with a decimal between 0 and 1.\nThis value can have a significant impact on the analysis,\nso explore how it impacts your data!\n\nMany calculations are dependent on knowing where certain metadata is stored.\nNamely, the experimental conditions (specified with ``--group``)\nthe compartment of origin (specified with ``--compartment``),\nthe sample ID (specified with ``--sample-id``),\nand the lipidomics mode (specified with ``--mode``).\nIf these are not passed,\nthen they default to "Group", "Compartment", "SampleID", and "Mode" respectively.\nTo find these rows,\nwe also need to know the number of lines in your column metadata.\nThis is specified with ``--n-rows-metadata``.\nPlease the section on [expected data file structure](data) for more information.\n\nThe error-normalised fold change (ENFC) calculation must know the labels for\nexperimental and control group.\nWithout this knowledge,\nthe concept of fold change is meaningless.\nTo specify, pass ``--control control``.\nEvery condition specified in ``Group`` will then be divided by ``control``\nto calculate the ENFC for all conditions.\n\n```shell\nlta data results --control lean\n```\n\n(configuration)=\n\n#### Configuration files\n\nIf you find yourself regularly passing arguments via the CLI,\nyou might want to try a configuration file!\nThis is a simple text file that stores options in a simple format:\n\n```shell\noption=value\n```\n\nBy default,\nLTA looks for ``lta_conf.txt`` in your current directory.\nHowever,\nyou can name this file whatever you want,\nand let LTA know where to find it,\nby passing the config flag like so:\n\n```shell\nlta -c path/to/your/config.txt data results\n```\n\nIf you specify an option in the configuration file,\nthat will override LTA\'s defaults,\nand specifying an option at the command line will override the configuration file!\nThe config file doesn\'t need do exist, however,\nand is just a bit of sugar.\n\n### The Output\n\n```{warning}\nRe-running the analysis overwrites existing results,\nso be sure to either back up your data,\nor pass a different output folder!\n```\n\nThe output folder will contain 2 files and 2 directories.\nFor each type of lipid, you should see the following:\n\n1. `switch_individual_lipid.csv` - a table of lipids and their A/B/U/N classification.\n1. `switch_lipid_classes.csv` - a table counting the frequency of each lipid class within the A/B/U/N classification.\n1. `jaccard` - the Jaccard similarity and p-value for each lipid class.\n1. `enfc` - a folder containing the ENFC results.\n\nWithin the ENFC folder,\nyou should see 2 files per group:\n\n1. `GROUP_by_CONTROL_individual_lipids.csv` - the ENFC results for each lipid.\n1. `GROUP_by_CONTROL_lipid_classes.csv` - the mean and St.Dev. of ENFC, grouped by lipid class.\n\nWithing the Jaccard folder,\nyou should see 1 file per group:\n\n1. `GROUP_by_CONTROL_jaccard_similarity.csv` - the Jaccard similarity and p-value\nfor each lipid class\n\nA few notes!\nFold change will **always** be {math}`group / control`.\nThe Jaccard similarities are calculated between conditions specified in ``--group``\nacross both compartments and lipid classes.\nThe p-values for these similarities are calculated using the method outlined by\n[N. Chung, et. al.][jaccard].\nFor ENFC,\nfold change is only meaningful if both values are non-0.\nWhere this is *not* true,\nNaN has been substituted,\nleaving an empty cell in the output.\n\n## Contributing\n\nOpen-source software is only open-source becaues of the excellent community,\nso we welcome any and all contributions!\nIf you think you have found a bug,\nplease log a report in our [issues][issues].\nIf you think you can fix a bug,\nor have an idea for a new feature,\nplease see our guide on [contributing](./contributing.md)\nfor more information on how to get started!\nWhile here,\nwe request that you follow our [code of conduct](./coc.md)\nto help maintain a welcoming,\nrespectful environment.\n\n## Future Developments\n\n- [x] Improve Github actions to use caching for poetry and Nox\n- [x] Allow for multiple ENFC calculations in 1 run\n- [x] Provide example configuration and data\n- [ ] Increase test coverage\n- [ ] Automate plotting\n\n## Citations\n\nIf you use LTA in your work,\nplease cite the following manuscripts:\n\n1. Furse, S., _et al._ Lipid Traffic Analysis reveals the impact of high paternal carbohydrate intake on offsprings’ lipid metabolism. _Commun Biol_ **4**, 163 (2021). [https://doi.org/10.1038/s42003-021-01686-1][paper_1]\n1. Furse, S.[^eq], Fernandez-Twinn, D.S.[^eq], _et al._ Lipid metabolism is dysregulated before, during and after pregnancy in a mouse model of gestational diabetes. _Int. J. Mol. Sci._ **22**, 7452 (2021). [https://doi.org/10.3390/ijms22147452][paper_2]\n1. Furse, S., _et al._ Paternal nutritional programming of lipid metabolism is propagated by sperm and seminal plasma. _Metabolomics_ **2022**, [https://doi.org/10.17863/CAM.79565][paper_3]\n1. Furse, S., _et al._ A mouse model of gestational diabetes shows dysregulated lipid metabolism post-weaning, after return to euglycaemia.  _Nutrition & Diabetes_ **2022**, [_In Press_][paper_4]\n\n[^eq]: These authors contributed equally to this work.\n\n[github]: https://github.com/IMS-Bio2Core-Facility/lta "LTA Source Code"\n[readthedocs]: http://lta.readthedocs.io/ "LTA Documentation"\n[pypi]: https://pypi.org/project/gtexquery/ "LTA PyPI Package"\n[realpython]: https://realpython.com/installing-python/ "RealPython Install Python"\n[psf]: https://wiki.python.org/moin/BeginnersGuide/Download "PSF Install Python"\n[pyenv]: https://github.com/pyenv/pyenv "PyEnv"\n[pipx]: https://pypa.github.io/pipx/ "pipx"\n[venv]: https://docs.python.org/3/tutorial/venv.html "Python venv"\n[tags]: https://github.com/IMS-Bio2Core-Facility/lta/releases "LTA releases"\n[issues]: https://github.com/IMS-Bio2Core-Facility/lta/issues "LTA issues"\n[examples]: https://github.com/IMS-Bio2Core-Facility/lta/tree/main/examples "Examples"\n[jaccard]: https://doi.org/10.1186/s12859-019-3118-5 "Jaccard Probabilities"\n[lta_v1]: https://doi.org/10.5281/zenodo.4309347 " LTA, R version 1"\n[paper_1]: https://www.nature.com/articles/s42003-021-01686-1 "LTA citation 1"\n[paper_2]: https://www.mdpi.com/1422-0067/22/14/7452 "LTA citation 2"\n[paper_3]: https://doi.org/10.17863/CAM.79565 "LTA citation 3"\n[paper_4]: https://www.repository.cam.ac.uk/handle/1810/332691 "LTA citation 4"\n',
    'author': 'rbpatt2019',
    'author_email': 'rb.patterson.cross@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/IMS-Bio2Core-Facility/LTA',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8.0,<3.11',
}


setup(**setup_kwargs)
